<?php

namespace Drupal\graphviz_filter\Plugin\Filter;

use Drupal\Component\Utility\Html;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\filter\Plugin\FilterBase;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\graphviz_filter\Plugin\FilterFormatPluginManager;
use Drupal\filter\FilterProcessResult;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\File\FileSystem;
use Drupal\Core\Link;
use Drupal\Core\Url;
use Drupal\file\Entity\File;

/**
 * @Filter(
 *   id = "filter_graphviz",
 *   title = @Translation("Graphviz (DOT) syntax"),
 *   description = @Translation("Interprets text as DOT syntax and returns a PNG image of the generated graph."),
 *   type = Drupal\filter\Plugin\FilterInterface::TYPE_MARKUP_LANGUAGE,
 *   settings = {
 *     "dot_path" = "",
 *     "commands" = \Drupal\graphviz_filter\Plugin\Filter\FilterGraphviz::SUPPORTED_COMMANDS,
 *     "pear_path" = ""
 *   }
 * )
 */
class FilterGraphviz extends FilterBase implements ContainerFactoryPluginInterface {

  /**
   * Regular expression for identifiying graphviz code.
   *
   * @var string
   */
  const REGEX = '/\[graphviz\](.*?)\[\/graphviz\]/si';

  /**
   * Supported graphviz commands.
   *
   * @var string
   */
  const SUPPORTED_COMMANDS = 'dot, neato, twopi, fdp, circo';

  /**
   * @var array
   */
  protected $arguments = [];

  /**
   * @var \Drupal\graphviz_filter\Plugin\FilterFormatPluginManager
   */
  protected $filterFormatPluginManager;

  /**
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected $renderer;

  /**
   * The Messenger service.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * The logger factory.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected $loggerFactory;

  /**
   * The file_system service.
   *
   * @var \Drupal\Core\File\FileSystem;
   */
  protected $fileSystem;

  /**
   * The current user service.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected $user;

  /**
   * FilterGraphviz constructor.
   *
   * @param array $configuration
   * @param $plugin_id
   * @param $plugin_definition
   * @param \Drupal\graphviz_filter\Plugin\FilterFormatPluginManager $filterFormatPluginManager
   * @param \Drupal\Core\Render\RendererInterface $renderer
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   * @param \Drupal\Core\File\FileSystem
   * @param \Drupal\Core\Session\AccountProxyInterface
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition,
                              FilterFormatPluginManager $filterFormatPluginManager,
                              RendererInterface $renderer,
                              MessengerInterface $messenger,
                              LoggerChannelFactoryInterface $logger_factory,
                              FileSystem $file_system,
                              AccountProxyInterface $user) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->arguments = $this->defaultArguments();

    $this->filterFormatPluginManager = $filterFormatPluginManager;
    $this->renderer = $renderer;
    $this->messenger = $messenger;
    $this->loggerFactory = $logger_factory;
    $this->fileSystem = $file_system;
    $this->user = $user;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('plugin.manager.graphviz_filter_format'),
      $container->get('renderer'),
      $container->get('messenger'),
      $container->get('logger.factory'),
      $container->get('file_system'),
      $container->get('current_user')
    );
  }

  /**
   * Get graphviz file directory path.
   *
   * @return string
   */
  public static function getDirectory() {
    return file_default_scheme() . '://graphviz';
  }

  /**
   * Default arguments.
   *
   * @return array
   */
  protected function defaultArguments() {
    return [
      'command' => [
        'description' => $this->t('Process the graph script using this Graphviz command (`dot`, `neato`, `twopi`, etc).'),
        'value' => 'dot',
      ],
      'formats' => [
        'description' => $this->t('Comma-separated list of desired output formats.'),
        'value' => 'png',
      ],
      'title' => [
        'description' => $this->t('Title of the output image.'),
        'value' => '',
      ],
      'link-output' => [
        'description' => $this->t('Output a link to the output image instead of embedding it (needs @title).'),
        'value' => FALSE,
      ],
      'link-input' => [
        'description' => $this->t('Add a link to the input DOT script (needs @title).'),
        'value' => FALSE,
      ],
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function process($text, $langcode) {
    $text = preg_replace_callback(self::REGEX, 'self::filter', $text);
    return new FilterProcessResult($text);
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $form['dot_path'] = array(
      '#type' => 'textfield',
      '#title' => $this->t('Path of Graphviz commands'),
      '#default_value' => $this->settings['dot_path'],
      '#description' => $this->t("Specify the base path of the Graphviz commands ('dot', 'neato', etc). Leave blank if these commands are to be found on the system PATH which, on this server, is set to:<br /><pre>@path</pre>",
        ['@path' => getenv('PATH')]),
    );

    $form['commands'] = array(
      '#type' => 'textfield',
      '#title' => $this->t('Supported Graphviz commands'),
      '#default_value' => $this->settings['commands'],
      '#description' => $this->t('Specify the Graphviz commands that can be invoked by the filter, separated by commas. This prevents users from inserting malicious commands in the Graphviz script.'),
    );

    $form['pear_path'] = array(
      '#type' => 'textfield',
      '#title' => $this->t('Path of PEAR::Image_GraphViz package'),
      '#default_value' => $this->settings['pear_path'],
      '#description' => $this->t("Specify the base path of the PEAR::Image_GraphViz package. The specified directory should contain the following subdirectory structure: Image/GraphViz.php. Leave blank if this package was installed via 'pear install' or if it is present in the PHP include_path setting. Note that this setting is only needed for sub-modules CCK Schema and Workflow Graph, rather than the core Graphviz Filter module. On this server, the PHP include_path is set to:<br/><pre>@path</pre>",
        ['@path' => ini_get('include_path')]),
    );

    return $form;
  }

  /**
   * Get arguments from formats.
   *
   * @param $formats
   *
   * @return array
   */
  protected function getArgumentsFromFormats($formats) {
    $arguments = [];

    foreach ($formats as $format) {
      if (!empty($format['arguments'])) {
        $arguments += $format['arguments'];
      }
    }

    return $arguments;
  }

  /**
   * Provides filtering of graphviz code and conversion into specified format.
   *
   * @param $matches
   *   Matches from preg_replace_callback().
   *
   * @return string
   *   Filtered HTML with graphiz code converted to specified format.
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   *
   * @see \Drupal\graphviz_filter\Plugin\Filter\FilterGraphviz::process().
   */
  protected function filter($matches) {
    $text = $matches[1];

    $formats = $this->filterFormatPluginManager->getDefinitions();
    $this->arguments += $this->getArgumentsFromFormats($formats);
    $this->parseCommentsForArguments($text, $this->arguments);

    if (!$this->isSupportedCommand($this->arguments['command']['value'])) {
      $message = $this->t("Graphviz command '%command' is not supported. Please choose one of (%supported) or edit the Graphviz filter settings.",
        ['%command' => $this->arguments['command']['value'], '%supported' => $this->getSupportedCommands()]);
      $this->messenger->addError($message);
      $this->loggerFactory->get('graphviz_filter')->error($message);
      return $text;
    }

    // Create temporary file with DOT script.
    $directory = self::getDirectory();
    $filename = hash('sha256', $text);
    $this->fileSystem->prepareDirectory($directory, FileSystem::CREATE_DIRECTORY|FileSystem::MODIFY_PERMISSIONS);
    $destination = $directory . '/' . $filename . '.dot';
    $file = file_save_data($text, $destination, FileSystem::EXISTS_REPLACE);

    // Render desired formats.
    $build = [
      '#theme' => 'container',
    ];
    foreach (explode(',', $this->arguments['formats']['value']) as $format) {
      $format = trim($format);
      $classes = [
        'graphviz',
        'graphviz-' . $format,
      ];
      $build['#attributes']['class'] = $classes;

      if (!array_key_exists($format, $formats)) {
        $message = $this->t('Graphviz format %format is not supported. Please choose one of (%supported).',
          ['%format' => $format, '%supported' => implode(', ', array_keys($formats))]);
        $this->messenger->addError($message);
        $this->loggerFactory->get('graphviz_filter')->error($message);
        continue;
      }

      $format_destination = $directory . '/' . $filename . '.' . $format;
      if (file_exists($format_destination)
          || self::render($destination, $format_destination, $format, $this->arguments['command']['value'])) {
        if ($this->arguments['link-output']['value']) {
          $attributes = [
            'class' => [
              'graphviz-link-output',
            ],
          ];
          $url = Url::fromUri('base:graphviz/' . $filename . '.' . $format, ['attributes' => $attributes]);
          $build['#children'] = [
            '#markup' => Link::fromTextAndUrl($this->t($this->arguments['title']['value']), $url)->toString(),
          ];
        }
        else {
          $build['#children'] = $this->filterFormatPluginManager
            ->createInstance($format)
            ->render($destination, $format_destination, $filename, $format, $this->arguments, $this);
        }

        if ($this->arguments['link-input']['value']) {
          $build['#children'][] = $this->buildLinkInput($format, $destination, $this->arguments);
        }
      }
    }

    return $this->renderer->render($build);
  }

  /**
   * Build link input.
   *
   * @param $format
   * @param $destination
   * @param array $arguments
   *
   * @return array
   */
  protected function buildLinkInput($format, $destination, array $arguments) {
    $attributes = [
      'class' => [
        'graphviz-link-input',
      ],
    ];

    return [
      '#theme' => 'container',
      '#attributes' => [
        'class' => [
          'graphviz',
          'graphviz-' . $format,
        ],
      ],
      '#children' => [
        '#markup' => Link::fromTextAndUrl($this->t('Download input script for %title',
          ['%title' => $arguments['title']['value']]), Url::fromUri($destination), ['attributes' => $attributes])->toString(),
      ],
    ];
  }

  /**
   * Check if command is supported.
   *
   * @param $command
   *
   * @return bool
   */
  protected function isSupportedCommand($command) {
    return in_array($command, $this->getSupportedCommandsArray());
  }

  /**
   * Get supported commands.
   *
   * @return string
   */
  protected function getSupportedCommands() {
    return (!empty($this->settings['commands'])) ? $this->settings['commands'] : self::SUPPORTED_COMMANDS;
  }

  /**
   * Get supported commands array.
   *
   * @return array
   */
  protected function getSupportedCommandsArray() {
    return array_map('trim', explode(',', $this->getSupportedCommands()));
  }

  /**
   * Create file path.
   *
   * @param $path
   * @param $filename
   *
   * @return string
   */
  public static function createFilePath($path, $filename) {
    if (!empty($path)) {
      return rtrim($path, '/') .'/'. $filename;
    }

    return $filename;
  }

  /**
   * Render graphviz format.
   *
   * @param $destination
   * @param $format_destination
   * @param $format
   * @param $command
   *
   * @return bool
   */
  public function render($destination, $format_destination, $format, $command) {
    $command = self::createFilePath('', $command);
    $command .= ' -T' . escapeshellarg($format);
    $command .= ' -o' . escapeshellarg($this->fileSystem->realpath($format_destination));
    $command .= ' ' . escapeshellarg($this->fileSystem->realpath($destination));
    $command .= ' 2>&1';

    $output = [];
    $return = 0;
    exec($command, $output, $return);
    clearstatcache();

    if ($return != 0) {
      $message = self::t('Graphviz encountered an error while rendering to format %format:<br />%output',
        ['%format' => $format, '%output' => Html::escape(implode('<br />', $output))]);
      $this->messenger->addError($message);
      $this->loggerFactory->get('graphviz_filter')->error($message);
      return FALSE;
    }

    $file = File::create([
      'uri' => $format_destination,
      'status' => 1,
      'uid' => $this->user->id(),
    ]);
    $file->save();

    return TRUE;
  }

  /**
   * Parse comments for arguments.
   *
   * @param string $text
   * @param array $arguments
   */
  protected function parseCommentsForArguments($text, array &$arguments) {
    // Parse comments in the DOT syntax to find our arguments.
    // Taken from a discussion in http://ostermiller.org/findcomment.html
    if (!preg_match_all("/(?:\/\*(?:[^*]|(?:\*+[^*\/]))*\*+\/)/", $text, $comments)) {
      return;
    }

    foreach ($comments[0] as $comment) {
      foreach ($arguments as $key => $info) {
        if (($match = $this->parseCommentForArgument($comment, $key)) !== FALSE) {
          $arguments[$key]['value'] = (is_bool($info['value']))
                                    ? preg_match("/TRUE|YES|ON|1/i", $match) == 1
                                    : trim($match);
        }
      }
    }
  }

  /**
   * Parse comment for argument.
   *
   * @param $comment
   * @param $argument
   *
   * @return bool|mixed
   */
  protected function parseCommentForArgument($comment, $argument) {
    // TODO This doesn't handle case /* @arg=value */ i.e. start and close on same line.
    if (preg_match("/@$argument\s*?=\s*?(\w[\w\W]*?)[\r\n]/i", $comment, $match)) {
      return $match[1];
    }

    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function tips($long = FALSE) {
    $output = '';

    if ($long) {
      $output .= $this->t('<h4>Using Graphviz filter</h4>');
    }

    $output .= $this->t('Use <a href=":dot" target="_blank">DOT syntax</a> with some additional commands embedded in the comments.',
      [':dot' => 'http://www.graphviz.org/documentation/']) . ' ';

    $formats = $this->filterFormatPluginManager->getDefinitions();
    $this->arguments += $this->getArgumentsFromFormats($formats);
    if ($long) {
      $output .= $this->t('In the comments, you can specify additional arguments such as:');
      $items = [
        '#theme' => 'item_list',
      ];
      foreach ($this->arguments as $argument => $info) {
        $items['#items'][] = '@'. $argument . ': ' . $info['description'];
      }
      $output .= $this->renderer->render($items);
    }

    $output .= $this->t('Available formats:');
    $items = [
      '#theme' => 'item_list',
    ];
    foreach ($formats as $format => $info) {
      $items['#items'][] = $format . ': ' . $info['description'];
    }
    $output .= $this->renderer->render($items);

    return $output;
  }

}
