<?php

namespace Drupal\graphviz_filter\Plugin\FilterFormat;

use Drupal\Component\Plugin\PluginBase;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\graphviz_filter\Plugin\FilterFormatInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Component\Utility\Html;
use Drupal\filter\FilterPluginManager;

/**
 * Base class for FilterFormat plugins.
 */
abstract class FilterFormatBase extends PluginBase implements FilterFormatInterface, ContainerFactoryPluginInterface {
  use StringTranslationTrait;

  /**
   * @var \Drupal\filter\FilterPluginManager
   */
  protected $filterPluginManager;

  /**
   * FilterFormatBase constructor.
   *
   * @param array $configuration
   * @param $plugin_id
   * @param $plugin_definition
   * @param \Drupal\filter\FilterPluginManager $filter_plugin_manager
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition,
                              FilterPluginManager $filter_plugin_manager) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->filterPluginManager = $filter_plugin_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('plugin.manager.filter')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function render($destination, $format_destination, $filename, $format, array $arguments) {
    $filter_graphviz = $this->filterPluginManager->createInstance('filter_graphviz');
    $build = [];

    $build['image'] = [
      '#theme' => 'image',
      '#uri' => $format_destination,
      '#attributes' => [
        'title' => Html::escape($this->t($arguments['title']['value'])),
      ],
    ];

    $uri = $format_destination . '.map';
    if ($arguments['imagemap']['value']) {
      if (file_exists($uri)
          || $filter_graphviz->render($destination, $uri, 'cmapx', $arguments['command']['value'])) {
        $build += $this->buildMapFromFile($uri);
        $build['image']['#attributes']['usemap'] = '#' . $build['map']['#attributes']['id'];
      }
    }

    return $build;
  }

  /**
   * Build map render array from file.
   *
   * @param string $uri
   *   Uri to file.
   *
   * @return array
   */
  protected function buildMapFromFile($uri) {
    $build['map'] = [
      '#type' => 'map_tag',
    ];

    $map = file_get_contents($uri);
    $id = (preg_match('/<map[^>]*id\s*=\s*"(.*?)"/', $map, $match))
      ? $match[1]
      : 'G';
    $build['map']['#attributes'] = [
      'id' => $id,
      'name' => $id,
    ];

    $dom = Html::load($map);
    $areas = $dom->getElementsByTagName('area');
    foreach ($areas as $area) {
      $attributes = [];
      if ($area->hasAttributes()) {
        foreach ($area->attributes as $name => $attrNode) {
          $attributes[$name] = $attrNode->value;
        }
      }

      $build['map']['#children'][] = [
        '#type' => 'html_tag',
        '#tag' => 'area',
        '#attributes' => $attributes,
      ];
    }

    return $build;
  }
}
